import json
import logging

import requests
from django.conf import settings

from unicef_vision.exceptions import VisionException

logger = logging.getLogger(__name__)

# VISION_NO_DATA_MESSAGE is what the remote vision system returns when it has no data
VISION_NO_DATA_MESSAGE = 'No Data Available'


class VisionDataLoader:
    """Base class for Data Loading"""
    URL = settings.VISION_URL

    def __init__(self, business_area_code=None, endpoint=None):
        if endpoint is None:
            raise VisionException('You must set the ENDPOINT name')

        separator = '' if self.URL.endswith('/') else '/'

        self.url = '{}{}{}'.format(self.URL, separator, endpoint)
        if business_area_code:
            self.url += '/{}'.format(business_area_code)

        logger.info('About to get data from {}'.format(self.url))

    def get(self):
        response = requests.get(
            self.url,
            headers={'Content-Type': 'application/json'},
            auth=(settings.VISION_USER, settings.VISION_PASSWORD),
            verify=False
        )

        if response.status_code != 200:
            raise VisionException('Load data failed! Http code: {}'.format(response.status_code))
        json_response = response.json()
        if json_response == VISION_NO_DATA_MESSAGE:
            return []

        return json_response


class ManualDataLoader(VisionDataLoader):
    """
    Can be used to sync single objects from VISION url templates:
    /endpoint if no business_area_code or object_number
    /endpoint/business_area_code if no object number provided
    /endpoint/object_number else
    """

    def __init__(self, business_area_code=None, endpoint=None, object_number=None):
        if not object_number:
            super().__init__(business_area_code=business_area_code, endpoint=endpoint)
        else:
            if endpoint is None:
                raise VisionException('You must set the ENDPOINT name')
            self.url = '{}/{}/{}'.format(
                self.URL,
                endpoint,
                object_number
            )


class FileDataLoader:
    """Loader to read json file instead of REST API"""

    def __init__(self, filename):
        self.filename = filename

    def get(self):
        data = json.load(open(self.filename))
        return data



# Generated by Django 2.1.5 on 2019-02-06 15:21

from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='VisionLog',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('handler_name', models.CharField(max_length=50, verbose_name='Handler Name')),
                ('business_area_code', models.CharField(blank=True, max_length=10, null=True, verbose_name='Business Area Code')),
                ('total_records', models.IntegerField(default=0, verbose_name='Total Records')),
                ('total_processed', models.IntegerField(default=0, verbose_name='Total Processed')),
                ('successful', models.BooleanField(default=False, verbose_name='Successful')),
                ('details', models.CharField(blank=True, default='', max_length=2048, verbose_name='Details')),
                ('exception_message', models.TextField(blank=True, default='', verbose_name='Exception Message')),
                ('date_processed', models.DateTimeField(auto_now=True, verbose_name='Date Processed')),
            ],
            options={
                'abstract': False,
            },
        ),
    ]



import json
import logging
import sys
import types
from abc import ABCMeta, abstractmethod
from collections import OrderedDict

from django.db.models import NOT_PROVIDED
from django.utils.encoding import force_text

from unicef_vision.exceptions import VisionException
from unicef_vision.loaders import FileDataLoader, ManualDataLoader, VisionDataLoader
from unicef_vision.utils import get_vision_logger_domain_model, wcf_json_date_as_datetime

logger = logging.getLogger(__name__)


class Empty:
    pass


class DataSynchronizer:

    __metaclass__ = ABCMeta

    REQUIRED_KEYS = {}
    GLOBAL_CALL = False
    LOADER_CLASS = None
    LOADER_EXTRA_KWARGS = []
    business_area_code = None

    @abstractmethod
    def _convert_records(self, records):    # pragma: no cover
        pass

    @abstractmethod
    def _save_records(self, records):   # pragma: no cover
        pass

    @abstractmethod
    def _get_kwargs(self):  # pragma: no cover
        return {}

    def __init__(self, business_area_code=None, *args, **kwargs) -> None:
        if not business_area_code:
            raise VisionException('business_area_code is required')
        self.business_area_code = business_area_code
        logger.info('Synchronizer is {}'.format(self.__class__.__name__))
        logger.info('business_area_code is {}'.format(business_area_code))

    def _filter_records(self, records):
        def is_valid_record(record):
            for key in self.REQUIRED_KEYS:
                if key not in record:
                    return False
            return True

        return [rec for rec in records if is_valid_record(rec)]

    def preload(self):  # pragma: no cover
        """hook to execute custom code before loading"""
        pass

    def logger_parameters(self):
        return {
            'handler_name': self.__class__.__name__,
            'business_area_code': self.business_area_code
        }

    def sync(self):
        """
        Performs the database sync
        :return:
        """
        self.log = get_vision_logger_domain_model()(**self.logger_parameters())

        loader_kwargs = self._get_kwargs()
        loader_kwargs.update({
            kwarg_name: getattr(self, kwarg_name)
            for kwarg_name in self.LOADER_EXTRA_KWARGS
        })
        data_getter = self.LOADER_CLASS(**loader_kwargs)

        try:
            original_records = data_getter.get()
            logger.info('{} records returned from get'.format(len(original_records)))

            converted_records = self._convert_records(original_records)
            self.log.total_records = len(converted_records)
            logger.info('{} records returned from conversion'.format(len(converted_records)))

            totals = self._save_records(converted_records)
        except Exception as e:
            logger.info('sync', exc_info=True)
            self.log.exception_message = force_text(e)
            traceback = sys.exc_info()[2]
            raise VisionException(force_text(e)).with_traceback(traceback)
        else:
            if isinstance(totals, dict):
                self.log.total_processed = totals.get('processed', 0)
                self.log.details = totals.get('details', '')
                self.log.total_records = totals.get('total_records', self.log.total_records)
            else:
                self.log.total_processed = totals
            self.log.successful = True
        finally:
            self.log.save()


class VisionDataSynchronizer(DataSynchronizer):
    __metaclass__ = ABCMeta

    ENDPOINT = None
    LOADER_CLASS = VisionDataLoader

    def __init__(self, business_area_code=None, *args, **kwargs) -> None:

        super().__init__(business_area_code, *args, **kwargs)

        if self.ENDPOINT is None:
            raise VisionException('You must set the ENDPOINT name')

    def _get_kwargs(self):
        return {
            'business_area_code': self.business_area_code,
            'endpoint': self.ENDPOINT,
        }


class FileDataSynchronizer(DataSynchronizer):
    __metaclass__ = ABCMeta

    LOADER_CLASS = FileDataLoader
    LOADER_EXTRA_KWARGS = ['filename', ]

    def __init__(self, business_area_code=None, *args, **kwargs):

        super().__init__(business_area_code, *args, **kwargs)
        filename = kwargs.get('filename', None)
        if not filename:
            raise VisionException('You need provide the path to the file')
        self.filename = filename


class MultiModelDataSynchronizer(VisionDataSynchronizer):
    MODEL_MAPPING = {}
    MAPPING = OrderedDict()
    DATE_FIELDS = []
    DEFAULTS = {}
    FIELD_HANDLERS = {}

    def _convert_records(self, records):
        if isinstance(records, list):
            return records
        try:
            return json.loads(records)
        except ValueError:
            return []

    def _get_field_value(self, field_name, field_json_code, json_item, model):
        if field_json_code in self.DATE_FIELDS:
            # parsing field as date
            return wcf_json_date_as_datetime(json_item[field_json_code])
        elif field_name in self.MODEL_MAPPING.keys():
            # this is related model, so we need to fetch somehow related object.
            related_model = self.MODEL_MAPPING[field_name]

            if isinstance(related_model, types.FunctionType):
                # callable provided, object should be returned from it
                result = related_model(data=json_item, key_field=field_json_code)
            else:
                # model class provided, related object can be fetched with query by field
                # analogue of field_json_code
                reversed_dict = dict(zip(
                    self.MAPPING[field_name].values(),
                    self.MAPPING[field_name].keys()
                ))
                result = related_model.objects.get(**{
                    reversed_dict[field_json_code]: json_item.get(field_json_code, None)
                })
        else:
            # field can be used as it is without custom mappings. if field has default, it should be used
            result = json_item.get(field_json_code, Empty)
            if result is Empty:
                # try to get default for field
                field_default = model._meta.get_field(field_name).default
                if field_default is not NOT_PROVIDED:
                    result = field_default

        # additional logic on field may be applied
        value_handler = self.FIELD_HANDLERS.get(
            {y: x for x, y in self.MODEL_MAPPING.items()}.get(model), {}
        ).get(field_name, None)
        if value_handler:
            result = value_handler(result)
        return result

    def _process_record(self, json_item):
        try:
            for model_name, model in self.MODEL_MAPPING.items():
                mapped_item = dict(
                    [(field_name, self._get_field_value(field_name, field_json_code, json_item, model))
                     for field_name, field_json_code in self.MAPPING[model_name].items()]
                )
                kwargs = dict(
                    [(field_name, value) for field_name, value in mapped_item.items()
                     if model._meta.get_field(field_name).unique]
                )

                if not kwargs:
                    for fields in model._meta.unique_together:
                        if all(field in mapped_item.keys() for field in fields):
                            unique_fields = fields
                            break

                    kwargs = {
                        field: mapped_item[field] for field in unique_fields
                    }

                defaults = dict(
                    [(field_name, value) for field_name, value in mapped_item.items()
                     if field_name not in kwargs.keys()]
                )
                defaults.update(self.DEFAULTS.get(model, {}))
                model.objects.update_or_create(
                    defaults=defaults, **kwargs
                )
        except Exception:
            logger.warning('Exception processing record', exc_info=True)

    def _save_records(self, records):
        processed = 0
        filtered_records = self._filter_records(records)

        for record in filtered_records:
            self._process_record(record)
            processed += 1
        return processed


class ManualVisionSynchronizer(MultiModelDataSynchronizer):
    LOADER_CLASS = ManualDataLoader
    LOADER_EXTRA_KWARGS = ['object_number', ]

    def __init__(self, business_area_code=None, object_number=None):
        self.object_number = object_number

        if not object_number:
            super().__init__(business_area_code=business_area_code)
        else:
            if self.ENDPOINT is None:
                raise VisionException('You must set the ENDPOINT name')

            self.business_area_code = business_area_code

            logger.info('Business area code is {}'.format(business_area_code))



import datetime
import json

import requests
from django.apps import apps
from django.conf import settings


def get_vision_logger_domain_model():
    get_model = apps.get_model
    return get_model(settings.VISION_LOGGER_MODEL)


def get_data_from_insight(endpoint, data={}):
    url = '{}/{}'.format(
        settings.VISION_URL,
        endpoint
    ).format(**data)

    response = requests.get(
        url,
        headers={'Content-Type': 'application/json'},
        auth=(settings.VISION_USER, settings.VISION_PASSWORD),
        verify=False
    )
    if response.status_code != 200:
        return False, 'Loading data from Vision Failed, status {}'.format(response.status_code)
    try:
        result = json.loads(response.json())
    except ValueError:
        return False, 'Loading data from Vision Failed, no valid response returned for data: {}'.format(data)
    return True, result


def wcf_json_date_as_datetime(jd):
    if jd is None:
        return None
    sign = jd[-7]
    if sign not in '-+' or len(jd) == 13:
        millisecs = int(jd[6:-2])
    else:
        millisecs = int(jd[6:-7])
        hh = int(jd[-7:-4])
        mm = int(jd[-4:-2])
        if sign == '-':
            mm = -mm
        millisecs += (hh * 60 + mm) * 60000
    return datetime.datetime(1970, 1, 1) \
        + datetime.timedelta(microseconds=millisecs * 1000)


def wcf_json_date_as_date(jd):
    if jd is None:
        return None
    sign = jd[-7]
    if sign not in '-+' or len(jd) == 13:
        millisecs = int(jd[6:-2])
    else:
        millisecs = int(jd[6:-7])
        hh = int(jd[-7:-4])
        mm = int(jd[-4:-2])
        if sign == '-':
            mm = -mm
        millisecs += (hh * 60 + mm) * 60000
    my_date = datetime.datetime(1970, 1, 1) + datetime.timedelta(microseconds=millisecs * 1000)
    return my_date.date()


def comp_decimals(y, x):
    def isclose(a, b, rel_tol=1e-09, abs_tol=0.0):
        return abs(a - b) <= max(rel_tol * max(abs(a), abs(b)), abs_tol)

    return isclose(float(x), float(y))



from django.contrib import admin


class VisionLoggerAdmin(admin.ModelAdmin):

    def has_add_permission(self, request):
        return False

    list_filter = (
        'handler_name',
        'successful',
        'date_processed',
    )
    list_display = (
        'handler_name',
        'total_records',
        'total_processed',
        'successful',
        'date_processed',
    )
    readonly_fields = (
        'details',
        'handler_name',
        'total_records',
        'total_processed',
        'successful',
        'exception_message',
        'date_processed',
    )



import argparse
import json
import logging
import os
from urllib.parse import urljoin

import requests
from django.conf import settings
from requests.auth import HTTPDigestAuth

logger = logging.getLogger(__name__)


class VisionAPIClient:
    """Client to Synchronize with Vision"""

    def __init__(self, username=None, password=None, base_url=settings.VISION_URL):
        self.base_url = base_url
        if username and password:
            self.auth = HTTPDigestAuth(username, password)

    def build_path(self, path=None):
        """ Builds the full path to the service.
        Args:
            path (string): The part of the path you want to append to the base url.
        Returns:
            A string containing the full path to the endpoint.
            e.g if the base_url was "http://woo.com" and the path was
            "databases" it would return "http://woo.com/databases/"
        """
        if path is None:
            return self.base_url
        return urljoin(
            self.base_url, os.path.normpath(path),
        )

    def make_request(self, path):

        response = requests.get(
            self.build_path(path),
            auth=getattr(self, 'auth', ()),
        )
        return response

    def call_command(self, command_type, **properties):

        payload = json.dumps(
            {
                'type': command_type,
                'command': {
                    'properties': properties
                }
            }
        )

        response = requests.post(
            self.build_path('command'),
            headers={'cache-control': 'application/json'},
            auth=getattr(self, 'auth', ()),
            data=payload,
        )
        return response


def main():
    """Main method for command line usage"""
    parser = argparse.ArgumentParser(description='VISION API Python Client')

    parser.add_argument('-U', '--username',
                        type=str,
                        default='',
                        help='Optional username for authentication')
    parser.add_argument('-P', '--password',
                        type=str,
                        default='',
                        help='Optional password for authentication')

    args = parser.parse_args()
    VisionAPIClient(username=args.username, password=args.password)


if __name__ == '__main__':  # pragma: no cover
    main()



from django.db import models
from django.utils.translation import ugettext_lazy as _


class AbstractVisionLog(models.Model):
    """Represents a sync log for Vision SAP"""

    handler_name = models.CharField(max_length=50, verbose_name=_('Handler Name'))
    business_area_code = models.CharField(max_length=10, verbose_name=_('Business Area Code'), null=True, blank=True)
    total_records = models.IntegerField(default=0, verbose_name=_('Total Records'))
    total_processed = models.IntegerField(default=0, verbose_name=_('Total Processed'))
    successful = models.BooleanField(default=False, verbose_name=_('Successful'))
    details = models.CharField(max_length=2048, blank=True, default='', verbose_name=_('Details'))
    exception_message = models.TextField(blank=True, default='', verbose_name=_('Exception Message'))
    date_processed = models.DateTimeField(auto_now=True, verbose_name=_('Date Processed'))

    def __str__(self):
        return u'{0.business_area_code}: {0.date_processed}:{0.successful} {0.total_processed}'.format(self)

    class Meta:
        abstract = True



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import ast
import codecs
import os
import re
import sys

from setuptools import find_packages, setup
from setuptools.command.install import install
from setuptools.command.test import test as TestCommand

HERE = os.path.abspath(os.path.dirname(__file__))
init = os.path.join(HERE, "src", "unicef_vision", "__init__.py")

_version_re = re.compile(r'__version__\s+=\s+(.*)')
_name_re = re.compile(r'NAME\s+=\s+(.*)')

with open(init, 'rb') as f:
    content = f.read().decode('utf-8')
    VERSION = str(ast.literal_eval(_version_re.search(content).group(1)))
    NAME = str(ast.literal_eval(_name_re.search(content).group(1)))


def read(*files):
    content = ''
    for f in files:
        content += codecs.open(os.path.join(HERE, f), 'r').read()
    return content


class VerifyTagVersion(install):
    """Verify that the git tag matches version"""

    def run(self):
        tag = os.getenv("CIRCLE_TAG")
        if tag != VERSION:
            info = "Git tag: {} does not match the version of this app: {}".format(
                tag,
                VERSION
            )
            sys.exit(info)


class PyTest(TestCommand):
    user_options = [('pytest-args=', 'a', "Arguments to pass to pytest")]

    def initialize_options(self):
        TestCommand.initialize_options(self)
        self.pytest_args = ''

    def run_tests(self):
        import shlex
        # import here, cause outside the eggs aren't loaded
        import pytest
        errno = pytest.main(shlex.split(self.pytest_args))
        sys.exit(errno)


setup(
    name=NAME,
    version=VERSION,
    url='https://github.com/unicef/unicef-vision',
    author='UNICEF',
    author_email='rapidpro@unicef.org',
    description='',
    long_description=read('README.rst'),
    platforms=['any'],
    license='Apache 2 License',
    classifiers=[
        'Environment :: Web Environment',
        'Framework :: Django',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
    ],
    install_requires=[
        'Django',
        'requests',
    ],
    extras_require={
        'test': [
            'responses',
            'mock',
            'factory-boy',
            'pytest',
            'pytest-pythonpath',
            'pytest-cov',
            'pytest-django',
            'pytest-echo',
            'isort',
            'flake8',
        ],
    },
    package_dir={'': 'src'},
    packages=find_packages('src'),
    include_package_data=True,
    cmdclass={"verify": VerifyTagVersion,
              'test': PyTest,
              }
)